#include <iostream>
#include <vector>
#include <omp.h>
#include <cmath>

#include "Utils.hpp"

using namespace std;

double computeVectorLength(const vector<double> u) {
    double sumSquares = 0;
    for (int i = 0; i < u.size(); i++) {
        sumSquares += u[i] * u[i];
    }

    return sqrt(sumSquares);
}

vector<double> normalizationSequential(vector<double> u) {
    double sumSquares = 0;
    for (int i = 0; i < u.size(); i++) {
        sumSquares += u[i] * u[i];
    }

    double vectorLength = sqrt(sumSquares);
    for (int i = 0; i < u.size(); i++) {
        u[i] /= vectorLength;
    }

    return u;
}

vector<double> normalizationParallelSections(vector<double> u) {
    int lastIndexInFirstHalf = (u.size() - 1) / 2;

    double leftHalfSumSquares = 0;
    double rightHalfSumSquares = 0;
    #pragma omp parallel num_threads(2)
    {
        #pragma omp sections
        {
            #pragma omp section
            {
                // QUESTION: what would happen if we wrote directly to leftHalfSumOfSquares?
                double halfSumSquares = 0.0;
                for (int i = 0; i <= lastIndexInFirstHalf; i++) {
                    halfSumSquares += u[i] * u[i];
                }
                leftHalfSumSquares = halfSumSquares;
            }

            #pragma omp section
            {
                double halfSumSquares = 0.0;
                for (int i = lastIndexInFirstHalf + 1; i < u.size(); i++) {
                    halfSumSquares += u[i] * u[i];
                }
                rightHalfSumSquares = halfSumSquares;
            }
        }

        double vectorLength = sqrt(leftHalfSumSquares + rightHalfSumSquares);

        #pragma omp sections
        {
            #pragma omp section
            {
                for (int i = 0; i <= lastIndexInFirstHalf; i++) {
                    u[i] /= vectorLength;
                }
            }

            #pragma omp section
            {
                double halfSumSquares = 0;
                for (int i = lastIndexInFirstHalf + 1; i < u.size(); i++) {
                    u[i] /= vectorLength;
                }
            }
        }
    }

    return u;
}

vector<double> normalizationParallelForAndCritical(vector<double> u) {
    double sumSquares = 0;
    #pragma omp parallel for
    for (int i = 0; i < u.size(); i++) {
        #pragma omp critical
        sumSquares += u[i] * u[i];
    }

    double vectorLength = sqrt(sumSquares);
    #pragma omp parallel for
    for (int i = 0; i < u.size(); i++) {
        u[i] /= vectorLength;
    }

    return u;
}

vector<double> normalizationParallelForAndReduction(vector<double> u) {
    double sumSquares = 0;
    #pragma omp parallel for reduction(+:sumSquares)
    for (int i = 0; i < u.size(); i++) {
        sumSquares += u[i] * u[i];
    }

    double vectorLength = sqrt(sumSquares);
    #pragma omp parallel for
    for (int i = 0; i < u.size(); i++) {
        u[i] /= vectorLength;
    }

    return u;
}

int main() {
    vector<double> u = generateRandomVector(50000000);

    cout << "Length of the input vector: " << computeVectorLength(u) << endl;

    {
        Stopwatch sw;
        sw.start();
        auto uNorm = normalizationSequential(u);
        sw.stop();
        cout << "Sequential: " << sw.duration().count() << " ms, length " << computeVectorLength(uNorm) << endl;
    }

    {
        Stopwatch sw;
        sw.start();
        auto uNorm = normalizationParallelSections(u);
        sw.stop();
        cout << "Parallel sections: " << sw.duration().count() << " ms, length " << computeVectorLength(uNorm) << endl;
    }

    {
        Stopwatch sw;
        sw.start();
        auto uNorm = normalizationParallelForAndCritical(u);
        sw.stop();
        cout << "Parallel for + critical section: " << sw.duration().count() << " ms, length " << computeVectorLength(uNorm) << endl;
    }

    {
        Stopwatch sw;
        sw.start();
        auto uNorm = normalizationParallelForAndReduction(u);
        sw.stop();
        cout << "Parallel for + reduction: " << sw.duration().count() << " ms, length " << computeVectorLength(uNorm) << endl;
    }

    return 0;
}
